/* 
    qasn1 - Asn1BerSerializer - serializer class for qasn1 library
    Copyright(C) 2010 - Andrius da Costa Ribas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Asn1BerSerializer.h"

class Asn1BerSerializerPrivate
{
public:
    Asn1BerSerializerPrivate(Asn1BerNode* node, QIODevice* oStream);
    Asn1BerSerializerPrivate(Asn1BerNode* node, QByteArray* oStream);
    
    QIODevice* m_outputStream;
    Asn1BerNode* m_subject;
};

Asn1BerSerializerPrivate::Asn1BerSerializerPrivate(Asn1BerNode* node, QIODevice* oStream)
{
    m_subject = node;
    m_outputStream = oStream;
}

Asn1BerSerializerPrivate::Asn1BerSerializerPrivate(Asn1BerNode* node, QByteArray* oStream)
{
    QBuffer* buffer = new QBuffer(oStream);
    buffer->open(QIODevice::WriteOnly);
    m_outputStream = buffer;
    m_subject = node;
}

Asn1BerSerializer::Asn1BerSerializer(Asn1BerNode* node, QIODevice* oStream) : d( new Asn1BerSerializerPrivate(node, oStream) )
{
}

Asn1BerSerializer::Asn1BerSerializer(Asn1BerNode* node, QByteArray* oStream) : d( new Asn1BerSerializerPrivate(node, oStream) )
{
}

Asn1BerSerializer::~Asn1BerSerializer()
{
    delete d;
}

void Asn1BerSerializer::serializeCurrentOnly()
{
    unsigned int tagNumber;
    char currentOctet;
    unsigned int length;
    QByteArray value;
    
    Q_ASSERT(d->m_outputStream->isWritable());
    
    if ( d->m_subject == 0 )
    {
        return;
    }
    
    /////// writting tag.
    
    // writting the class bits
    currentOctet = (unsigned char) d->m_subject->tagClass();
    
    // writting Primitive/Constructed bit
    if ( d->m_subject->isCodingComplex() )
    {
        currentOctet |= (unsigned char) CODING_COMPLEX;
    }
    
    tagNumber = d->m_subject->tag();
    if ( tagNumber < BIG_TAG_IND )
    {
        currentOctet |= (unsigned char) tagNumber;
        d->m_outputStream->putChar(currentOctet);
    }
    else
    {
        QByteArray tagNumberAux;
        bool first;
        
        first = true;
        
        currentOctet |= (unsigned char) BIG_TAG_IND;
        d->m_outputStream->putChar(currentOctet);
        do
        {
            // constructing the tag octets backwards
            if ( first )
            {
                currentOctet = (unsigned char) tagNumber % 0x80;
                tagNumber /= 0x80; // 7 bits each octet
                first = false;
            }
            else
            {
                currentOctet = (unsigned char) tagNumber % 0x80;
                currentOctet |= (unsigned char) BIG_TAG_NOT_LAST;
                tagNumber /= 0x80;
            }
            tagNumberAux.prepend(currentOctet);
        }while(tagNumber != 0);
        d->m_outputStream->write(tagNumberAux);
    }
    
    /////// writting length
    
    if ( !(d->m_subject->hasDefinedLength()) )
    {
        currentOctet = (unsigned char) UNKNOWN_LENGTH;
        d->m_outputStream->putChar(currentOctet);
    }
    else
    {
        length = d->m_subject->len();
        if ( length < BIG_LEN_IND )
        {
            currentOctet = (unsigned char) length;
            d->m_outputStream->putChar(currentOctet);
        }
        else
        {
            int lengthByteCount;
            QByteArray  bigLenAux;
            
            lengthByteCount = 0;
            
            do
            {
                lengthByteCount++;
                currentOctet = (unsigned char) (length % 0x100);
                length /= 0x100;
                bigLenAux.prepend(currentOctet);
            }while(length != 0);
            
            // TODO: error if lengthByteCount > 1111110(2)
            
            currentOctet = (unsigned char) lengthByteCount;
            currentOctet |= (unsigned char) BIG_LEN_IND;
            
            bigLenAux.prepend(currentOctet);
            
            d->m_outputStream->write(bigLenAux);
            
        }
    }
    
    /////// writting value
    
    if ( d->m_subject->isCodingPrimitive() )
    {
        d->m_outputStream->write(d->m_subject->val());
    }
    else
    {
        Asn1BerSerializer(d->m_subject->firstChild(), d->m_outputStream).serializeFromCurrent();
        if ( !(d->m_subject->hasDefinedLength()) )
        {
            d->m_outputStream->putChar(0);
            d->m_outputStream->putChar(0);
        }
    }
    
}

void Asn1BerSerializer::serializeFromCurrent()
{
    if ( d->m_subject != 0 )
    {
        serializeCurrentOnly();
        Asn1BerSerializer( d->m_subject->next(), d->m_outputStream ).serializeFromCurrent();
    }
}

void Asn1BerSerializer::serializeAllTree()
{
    if ( d->m_subject != 0 )
    {
        Asn1BerSerializer( d->m_subject->first(), d->m_outputStream ).serializeFromCurrent();
    }
}

